use crate::{schema, token};
use microrm::prelude::*;
use ring::signature::KeyPair;
use serde::{Deserialize, Serialize};

type Request = tide::Request<super::ServerStateWrapper>;

#[derive(serde::Serialize)]
pub enum OIDCErrorType {
    InvalidRequest,
    UnauthorizedClient,
    AccessDenied,
    UnsupportedResponseType,
    InvalidScope,
    ServerError,
    TemporarilyUnavailable,
}

/// error type,
pub struct OIDCError<'a>(OIDCErrorType, String, Option<&'a str>);

impl<'a> OIDCError<'a> {
    fn to_response(self) -> tide::Response {
        #[derive(Serialize)]
        struct ErrorOut<'a> {
            error: OIDCErrorType,
            error_description: String,
            state: Option<&'a str>,
        }

        let eo = ErrorOut {
            error: self.0,
            error_description: self.1,
            state: self.2,
        };

        tide::Response::builder(400)
            .body(serde_json::to_vec(&eo).unwrap())
            .build()
    }
}

#[derive(Deserialize)]
struct AuthorizeQueryParams {
    response_type: String,
    client_id: String,
    redirect_uri: String,
    scope: Option<String>,
}

fn do_code_authorize(
    request: Request,
    qp: AuthorizeQueryParams,
    state: Option<&str>,
    client: microrm::WithID<schema::Client>,
) -> Result<tide::Response, OIDCError> {
    todo!()
}

fn do_token_authorize(
    request: Request,
    qp: AuthorizeQueryParams,
    state: Option<&str>,
    client: microrm::WithID<schema::Client>,
) -> Result<tide::Response, OIDCError> {
    let qi = request.state().core.pool.query_interface();

    let shelper = super::session::SessionHelper::new(&request);

    let sauth = shelper
        .get_session(&request)
        .and_then(|sid| shelper.get_auth_for_session(shelper.get_realm().unwrap(), sid));

    if sauth.is_none() {
        // if we don't have any relevant auth info, redirect to login
        let mut login_url = request.url().join("../session/login").unwrap();
        login_url
            .query_pairs_mut()
            .clear()
            .append_pair("redirect", request.url().as_str());
        return Ok(tide::Redirect::new(login_url).into());
    }
    let sauth = sauth.unwrap();

    let scopes = qp
        .scope
        .as_ref()
        .map(|slist| slist.as_str())
        .unwrap_or("")
        .split_whitespace();

    let token = token::generate_auth_token(
        &request.state().core.config,
        &qi,
        shelper.get_realm().unwrap(),
        client.id(),
        sauth.user,
        scopes,
    );

    if let Some(token) = token {
        let response_body = serde_json::json!({
            "token": token,
        });
        Ok(tide::Response::builder(200)
            .content_type(tide::http::mime::JSON)
            .body(response_body)
            .build())
    } else {
        Err(OIDCError(
            OIDCErrorType::ServerError,
            "internal error while generating token".to_string(),
            state,
        ))
    }
}

fn do_authorize(request: Request, state: Option<&str>) -> Result<tide::Response, OIDCError> {
    let qp: AuthorizeQueryParams = request
        .query()
        .map_err(|x| OIDCError(OIDCErrorType::InvalidRequest, x.to_string(), state))?;

    // verify the realm and client_id and redirect_uri
    let qi = request.state().core.pool.query_interface();
    let realm = qi
        .get()
        .by(schema::Realm::Shortname, &request.param("realm").unwrap())
        .one()
        .expect("couldn't query db");
    if realm.is_none() {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "No such realm!".to_string(),
            state,
        ));
    }
    let realm = realm.unwrap();

    let client = qi
        .get()
        .by(schema::Client::Realm, &realm.id())
        .by(schema::Client::Shortname, &qp.client_id)
        .one()
        .expect("couldn't query db");
    if client.is_none() {
        return Err(OIDCError(
            OIDCErrorType::UnauthorizedClient,
            "Client does not exist".to_string(),
            state,
        ));
    }

    if qp.response_type == "code" {
        do_code_authorize(request, qp, state, client.unwrap())
    } else if qp.response_type == "token" {
        do_token_authorize(request, qp, state, client.unwrap())
    } else {
        return Err(OIDCError(
            OIDCErrorType::UnsupportedResponseType,
            "Only code and token are understood.".to_string(),
            state,
        ));
    }
}

async fn authorize(request: Request) -> tide::Result<tide::Response> {
    #[derive(Deserialize)]
    struct State {
        state: Option<String>,
    }
    let state: Option<String> = request.query::<State>().ok().map(|x| x.state).flatten();

    match do_authorize(request, state.as_ref().map(|x| x.as_str())) {
        Ok(r) => Ok(r),
        Err(e) => Ok(e.to_response()),
    }
}

#[derive(Deserialize)]
struct TokenRequestBody {
    grant_type: String,
    refresh_token: Option<String>,
    scope: Option<String>,
    redirect_uri: Option<String>,
    code: Option<String>,
    client_id: Option<String>,

    // direct grant
    username: Option<String>,
    password: Option<String>,
}

async fn token(mut request: Request) -> tide::Result<tide::Response> {
    let body: Result<TokenRequestBody, _> = request.body_form().await;
    // let body : TokenRequestBody = request.body_form();
    todo!()
}

const AUTHORIZE_PATH: &'static str = "oidc/authorize";
const TOKEN_PATH: &'static str = "oidc/token";
const JWKS_PATH: &'static str = "oidc/jwks";

async fn jwks(request: Request) -> tide::Result<tide::Response> {
    let qi = request.state().core.pool.query_interface();

    let shelper = super::session::SessionHelper::new(&request);
    let realm = shelper.get_realm()?;

    let keyinfo = qi
        .get()
        .by(schema::Key::Realm, &realm)
        .all()
        .expect("couldn't query db")
        .into_iter()
        .map(|key| {
            let kpair = ring::signature::Ed25519KeyPair::from_pkcs8(&key.keydata)
                .expect("couldn't parse keypair in db");
            let pubkey_bytes = kpair.public_key().as_ref();
            assert_eq!(pubkey_bytes.len(), 32);

            serde_json::json!({
                "kty": "OKP",
                "crv": "Ed25519",
                "x": base64::encode(pubkey_bytes),
                "kid": key.key_id,
            })
        });

    let jwks_response = serde_json::json!({
        "keys": keyinfo.collect::<Vec<_>>(),
    });

    Ok(tide::Response::builder(200).body(jwks_response).build())
}

async fn discovery_config(request: Request) -> tide::Result<tide::Response> {
    let server_config = &request.state().core.config;
    let base_url = format!(
        "{}/{}",
        server_config.base_url,
        request.param("realm").unwrap()
    );

    let config_response = serde_json::json!({
        "issuer": base_url,
        "authorization_endpoint": format!("{}/{}", base_url, AUTHORIZE_PATH),
        "token_endpoint": format!("{}/{}", base_url, TOKEN_PATH),
        "jwks_uri": format!("{}/{}", base_url, JWKS_PATH),
        "token_endpoint_auth_signing_alg_values_supported": ["EdDSA"],
        "response_types_supported": ["code", "id_token", "token id_token"],
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": ["EdDSA"],
    });

    Ok(tide::Response::builder(200).body(config_response).build())
}

pub(super) fn oidc_server(mut route: tide::Route<super::ServerStateWrapper>) {
    route.at(AUTHORIZE_PATH).get(authorize).post(authorize);
    route.at(TOKEN_PATH).post(token);
    route.at(JWKS_PATH).get(jwks);
    route
        .at(".well-known/openid-configuration")
        .get(discovery_config);
}
