use crate::{config, jwt, schema};
use microrm::{entity::EntityID, prelude::*};

pub fn generate_auth_token<'a>(
    config: &config::Config,
    qi: &microrm::QueryInterface,
    realm: schema::RealmID,
    client: schema::ClientID,
    user: schema::UserID,
    scopes: impl Iterator<Item = &'a str>,
) -> Option<String> {
    let realm = qi.get().by_id(&realm).one().ok()??;
    let client = qi.get().by_id(&client).one().ok()??;
    let user = qi.get().by_id(&user).one().ok()??;

    let issuer = format!("{}/{}", config.base_url, realm.shortname,);

    let iat = std::time::SystemTime::now();
    let exp = iat + std::time::Duration::from_secs(config.auth_token_expiry);

    // find all roles the user can possibly have access to
    let mut user_roles = vec![];
    for group in qi
        .get()
        .by(schema::GroupMembership::User, &user.id())
        .all()
        .ok()?
    {
        for group_role in qi
            .get()
            .by(schema::GroupRole::Group, &group.id())
            .all()
            .ok()?
        {
            user_roles.push(group_role.role);
        }
    }

    // find all roles requested by the scopes
    let mut requested_roles = vec![];
    for scope in scopes {
        if let Some(scope) = qi.get().by(schema::Scope::Shortname, scope).one().ok()? {
            for scope_role in qi
                .get()
                .by(schema::ScopeRole::Scope, &scope.id())
                .all()
                .ok()?
            {
                requested_roles.push(scope_role.role);
            }
        }
    }

    user_roles.sort_by_key(|k| k.raw_id());
    user_roles.dedup();
    requested_roles.sort_by_key(|k| k.raw_id());
    requested_roles.dedup();

    // find the intersection between requested roles and the ones the user actually has
    let resulting_roles = requested_roles
        .iter()
        .filter(|req| user_roles.contains(req))
        .map(|role_id| {
            serde_json::Value::String(
                qi.get()
                    .by_id(role_id)
                    .one()
                    .unwrap()
                    .unwrap()
                    .shortname
                    .clone(),
            )
        });

    let token = jwt::JWTData {
        sub: user.username.as_str(),
        iss: issuer.as_str(),
        aud: client.shortname.as_str(),
        iat: iat.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        exp: exp.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        extras: [(
            "roles",
            serde_json::Value::Array(resulting_roles.collect::<Vec<_>>()),
        )]
        .into(),
    };

    let key = qi.get().by(schema::Key::Realm, &realm.id()).one().ok()??;
    let kpair = ring::signature::Ed25519KeyPair::from_pkcs8(key.keydata.as_slice()).ok()?;

    Some(jwt::JWT::sign(&kpair, token).into_string())
}
