use crate::schema;
use microrm::prelude::*;

pub struct User {
    id: schema::UserID,
    model: Option<schema::User>,
}

static PBKDF2_ROUNDS: std::num::NonZeroU32 = unsafe { std::num::NonZeroU32::new_unchecked(20000) };

impl User {
    pub fn from_model(model: microrm::WithID<schema::User>) -> Self {
        Self {
            id: model.id(),
            model: Some(model.wrapped()),
        }
    }

    /// returns Some(true) if challenge passed, Some(false) if challenge failed, and None if
    /// challenge not found
    pub fn verify_challenge(
        &self,
        qi: &microrm::QueryInterface,
        which: schema::AuthChallengeType,
        response: &[u8],
    ) -> Option<bool> {
        let challenge = qi
            .get()
            .by(schema::AuthChallenge::User, &self.id)
            .by(schema::AuthChallenge::ChallengeType, &which)
            .one()
            .expect("couldn't query db")?;

        match which {
            schema::AuthChallengeType::Password => {
                self.verify_password_challenge(challenge.wrapped(), response)
            }
            schema::AuthChallengeType::TOTP => {
                self.verify_totp_challenge(challenge.wrapped(), response)
            }
            _ => todo!(),
        }
    }

    pub fn set_new_password(&self, qi: &microrm::QueryInterface, password: &[u8]) {
        qi.delete()
            .by(schema::AuthChallenge::User, &self.id)
            .by(
                schema::AuthChallenge::ChallengeType,
                &schema::AuthChallengeType::Password,
            )
            .exec()
            .expect("couldn't query db");

        let rng = ring::rand::SystemRandom::new();
        let salt: [u8; 16] = ring::rand::generate(&rng)
            .expect("Couldn't generate random salt?")
            .expose();

        let mut generated = [0u8; ring::digest::SHA256_OUTPUT_LEN];

        ring::pbkdf2::derive(
            ring::pbkdf2::PBKDF2_HMAC_SHA256,
            PBKDF2_ROUNDS,
            &salt,
            password,
            &mut generated,
        );

        qi.add(&schema::AuthChallenge {
            user: self.id,
            challenge_type: schema::AuthChallengeType::Password,
            public: salt.into(),
            secret: generated.into(),
        })
        .expect("couldn't set password");
    }

    fn verify_password_challenge(
        &self,
        challenge: schema::AuthChallenge,
        response: &[u8],
    ) -> Option<bool> {
        use ring::pbkdf2;

        Some(
            pbkdf2::verify(
                pbkdf2::PBKDF2_HMAC_SHA256,
                PBKDF2_ROUNDS,
                challenge.public.as_slice(),
                response,
                challenge.secret.as_slice(),
            )
            .is_ok(),
        )
    }

    fn verify_totp_challenge(
        &self,
        challenge: schema::AuthChallenge,
        response: &[u8],
    ) -> Option<bool> {
        todo!()
    }
}
